'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var _classCallCheck = _interopDefault(require('@babel/runtime/helpers/classCallCheck'));
var _createClass = _interopDefault(require('@babel/runtime/helpers/createClass'));
var _defineProperty = _interopDefault(require('@babel/runtime/helpers/defineProperty'));
var nacl = _interopDefault(require('tweetnacl'));
var BN = _interopDefault(require('bn.js'));
var bs58 = _interopDefault(require('bs58'));
var _regeneratorRuntime = _interopDefault(require('@babel/runtime/regenerator'));
var _asyncToGenerator = _interopDefault(require('@babel/runtime/helpers/asyncToGenerator'));
var BufferLayout = require('buffer-layout');
var _toConsumableArray = _interopDefault(require('@babel/runtime/helpers/toConsumableArray'));
var assert = _interopDefault(require('assert'));
var url = require('url');
var fetch = _interopDefault(require('node-fetch'));
var jayson = _interopDefault(require('jayson/lib/client/browser'));
var superstruct = require('superstruct');
var rpcWebsockets = require('rpc-websockets');
var _possibleConstructorReturn = _interopDefault(require('@babel/runtime/helpers/possibleConstructorReturn'));
var _getPrototypeOf = _interopDefault(require('@babel/runtime/helpers/getPrototypeOf'));
var _get = _interopDefault(require('@babel/runtime/helpers/get'));
var _inherits = _interopDefault(require('@babel/runtime/helpers/inherits'));

/**
 * A public key
 */

var PublicKey =
/*#__PURE__*/
function () {
  /**
   * Create a new PublicKey object
   */
  function PublicKey(value) {
    _classCallCheck(this, PublicKey);

    _defineProperty(this, "_bn", void 0);

    if (typeof value === 'string') {
      // hexadecimal number
      if (value.startsWith('0x')) {
        this._bn = new BN(value.substring(2), 16);
      } else {
        // assume base 58 encoding by default
        this._bn = new BN(bs58.decode(value));
      }
    } else {
      this._bn = new BN(value);
    }

    if (this._bn.byteLength() > 32) {
      throw new Error("Invalid public key input");
    }
  }
  /**
   * Checks if the provided object is a PublicKey
   */


  _createClass(PublicKey, [{
    key: "equals",

    /**
     * Checks if two publicKeys are equal
     */
    value: function equals(publicKey) {
      return this._bn.eq(publicKey._bn);
    }
    /**
     * Return the base-58 representation of the public key
     */

  }, {
    key: "toBase58",
    value: function toBase58() {
      return bs58.encode(this.toBuffer());
    }
    /**
     * Return the Buffer representation of the public key
     */

  }, {
    key: "toBuffer",
    value: function toBuffer() {
      var b = this._bn.toArrayLike(Buffer);

      if (b.length === 32) {
        return b;
      }

      var zeroPad = Buffer.alloc(32);
      b.copy(zeroPad, 32 - b.length);
      return zeroPad;
    }
    /**
     * Returns a string representation of the public key
     */

  }, {
    key: "toString",
    value: function toString() {
      return this.toBase58();
    }
  }], [{
    key: "isPublicKey",
    value: function isPublicKey(o) {
      return o instanceof PublicKey;
    }
  }]);

  return PublicKey;
}();

/**
 * An account key pair (public and secret keys).
 */

var Account =
/*#__PURE__*/
function () {
  /**
   * Create a new Account object
   *
   * If the secretKey parameter is not provided a new key pair is randomly
   * created for the account
   *
   * @param secretKey Secret key for the account
   */
  function Account() {
    var secretKey = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;

    _classCallCheck(this, Account);

    _defineProperty(this, "_keypair", void 0);

    if (secretKey) {
      this._keypair = nacl.sign.keyPair.fromSecretKey(secretKey);
    } else {
      this._keypair = nacl.sign.keyPair();
    }
  }
  /**
   * The public key for this account
   */


  _createClass(Account, [{
    key: "publicKey",
    get: function get() {
      return new PublicKey(this._keypair.publicKey);
    }
    /**
     * The **unencrypted** secret key for this account
     */

  }, {
    key: "secretKey",
    get: function get() {
      return this._keypair.secretKey;
    }
  }]);

  return Account;
}();

// These constants should match the values in
// https://github.com/solana-labs/solana/blob/master/sdk/src/timing.rs
var NUM_TICKS_PER_SECOND = 10;
var DEFAULT_TICKS_PER_SLOT = 8;

/**
 * Layout for a public key
 */

var publicKey = function publicKey() {
  var property = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 'publicKey';
  return BufferLayout.blob(32, property);
};
/**
 * Layout for a 64bit unsigned value
 */

var uint64 = function uint64() {
  var property = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 'uint64';
  return BufferLayout.blob(8, property);
};
/**
 * Layout for a Rust String type
 */

var rustString = function rustString() {
  var property = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 'string';
  var rsl = BufferLayout.struct([BufferLayout.u32('length'), BufferLayout.u32('lengthPadding'), BufferLayout.blob(BufferLayout.offset(BufferLayout.u32(), -8), 'chars')], property);

  var _decode = rsl.decode.bind(rsl);

  var _encode = rsl.encode.bind(rsl);

  rsl.decode = function (buffer, offset) {
    var data = _decode(buffer, offset);

    return data.chars.toString('utf8');
  };

  rsl.encode = function (str, buffer, offset) {
    var data = {
      chars: Buffer.from(str, 'utf8')
    };
    return _encode(data, buffer, offset);
  };

  return rsl;
};

function decodeLength(bytes) {
  var len = 0;
  var size = 0;

  for (;;) {
    var elem = bytes.shift();
    len |= (elem & 0x7f) << size * 7;
    size += 1;

    if ((elem & 0x80) === 0) {
      break;
    }
  }

  return len;
}
function encodeLength(bytes, len) {
  var rem_len = len;

  for (;;) {
    var elem = rem_len & 0x7f;
    rem_len >>= 7;

    if (rem_len == 0) {
      bytes.push(elem);
      break;
    } else {
      elem |= 0x80;
      bytes.push(elem);
    }
  }
}

/**
 * Maximum over-the-wire size of a Transaction
 */
var PACKET_DATA_SIZE = 512;
/**
 * List of TransactionInstruction object fields that may be initialized at construction
 *
 * @typedef {Object} TransactionInstructionCtorFields
 * @property {?Array<PublicKey>} keys
 * @property {?PublicKey} programId
 * @property {?Buffer} data
 */

/**
 * Transaction Instruction class
 */
var TransactionInstruction =
/**
 * Public keys to include in this transaction
 * Boolean represents whether this pubkey needs to sign the transaction
 */

/**
 * Program Id to execute
 */

/**
 * Program input
 */
function TransactionInstruction(opts) {
  _classCallCheck(this, TransactionInstruction);

  _defineProperty(this, "keys", []);

  _defineProperty(this, "programId", void 0);

  _defineProperty(this, "data", Buffer.alloc(0));

  opts && Object.assign(this, opts);
};
/**
 * @private
 */

/**
 * Transaction class
 */
var Transaction =
/*#__PURE__*/
function () {
  _createClass(Transaction, [{
    key: "signature",

    /**
     * Signatures for the transaction.  Typically created by invoking the
     * `sign()` method
     */

    /**
     * The first (primary) Transaction signature
     */
    get: function get() {
      if (this.signatures.length > 0) {
        return this.signatures[0].signature;
      }

      return null;
    }
    /**
     * The instructions to atomically execute
     */

  }]);

  /**
   * Construct an empty Transaction
   */
  function Transaction(opts) {
    _classCallCheck(this, Transaction);

    _defineProperty(this, "signatures", []);

    _defineProperty(this, "instructions", []);

    _defineProperty(this, "recentBlockhash", void 0);

    opts && Object.assign(this, opts);
  }
  /**
   * Add one or more instructions to this Transaction
   */


  _createClass(Transaction, [{
    key: "add",
    value: function add() {
      var _this = this;

      for (var _len = arguments.length, items = new Array(_len), _key = 0; _key < _len; _key++) {
        items[_key] = arguments[_key];
      }

      if (items.length === 0) {
        throw new Error('No instructions');
      }

      items.forEach(function (item) {
        if (item instanceof Transaction) {
          _this.instructions = _this.instructions.concat(item.instructions);
        } else {
          _this.instructions.push(new TransactionInstruction(item));
        }
      });
      return this;
    }
    /**
     * @private
     */

  }, {
    key: "_getSignData",
    value: function _getSignData() {
      var recentBlockhash = this.recentBlockhash;

      if (!recentBlockhash) {
        throw new Error('Transaction recentBlockhash required');
      }

      if (this.instructions.length < 1) {
        throw new Error('No instructions provided');
      }

      var keys = this.signatures.map(function (_ref) {
        var publicKey = _ref.publicKey;
        return publicKey.toString();
      });
      var numRequiredSignatures = 0;
      var programIds = [];
      this.instructions.forEach(function (instruction) {
        var programId = instruction.programId.toString();

        if (!programIds.includes(programId)) {
          programIds.push(programId);
        }

        instruction.keys.forEach(function (keySignerPair) {
          var keyStr = keySignerPair.pubkey.toString();

          if (keySignerPair.isSigner) {
            numRequiredSignatures += 1;
          }

          if (!keys.includes(keyStr)) {
            keys.push(keyStr);
          }
        });
      });
      var keyCount = [];
      encodeLength(keyCount, keys.length);
      var programIdCount = [];
      encodeLength(programIdCount, programIds.length);
      var instructions = this.instructions.map(function (instruction) {
        var data = instruction.data,
            programId = instruction.programId;
        var keyIndicesCount = [];
        encodeLength(keyIndicesCount, instruction.keys.length);
        var dataCount = [];
        encodeLength(dataCount, instruction.data.length);
        return {
          programIdIndex: programIds.indexOf(programId.toString()),
          keyIndicesCount: Buffer.from(keyIndicesCount),
          keyIndices: Buffer.from(instruction.keys.map(function (keyObj) {
            return keys.indexOf(keyObj.pubkey.toString());
          })),
          dataLength: Buffer.from(dataCount),
          data: data
        };
      });
      instructions.forEach(function (instruction) {
        assert(instruction.programIdIndex >= 0);
        instruction.keyIndices.forEach(function (keyIndex) {
          return assert(keyIndex >= 0);
        });
      });
      var instructionCount = [];
      encodeLength(instructionCount, instructions.length);
      var instructionBuffer = Buffer.alloc(PACKET_DATA_SIZE);
      Buffer.from(instructionCount).copy(instructionBuffer);
      var instructionBufferLength = instructionCount.length;
      instructions.forEach(function (instruction) {
        var instructionLayout = BufferLayout.struct([BufferLayout.u8('programIdIndex'), BufferLayout.blob(instruction.keyIndicesCount.length, 'keyIndicesCount'), BufferLayout.seq(BufferLayout.u8('keyIndex'), instruction.keyIndices.length, 'keyIndices'), BufferLayout.blob(instruction.dataLength.length, 'dataLength'), BufferLayout.seq(BufferLayout.u8('userdatum'), instruction.data.length, 'data')]);
        var length = instructionLayout.encode(instruction, instructionBuffer, instructionBufferLength);
        instructionBufferLength += length;
      });
      instructionBuffer = instructionBuffer.slice(0, instructionBufferLength);
      var signDataLayout = BufferLayout.struct([BufferLayout.blob(1, 'numRequiredSignatures'), BufferLayout.blob(keyCount.length, 'keyCount'), BufferLayout.seq(publicKey('key'), keys.length, 'keys'), publicKey('recentBlockhash'), BufferLayout.blob(programIdCount.length, 'programIdCount'), BufferLayout.seq(publicKey('programId'), programIds.length, 'programIds')]);
      var transaction = {
        numRequiredSignatures: Buffer.from([numRequiredSignatures]),
        keyCount: Buffer.from(keyCount),
        keys: keys.map(function (key) {
          return new PublicKey(key).toBuffer();
        }),
        recentBlockhash: Buffer.from(bs58.decode(recentBlockhash)),
        programIdCount: Buffer.from(programIdCount),
        programIds: programIds.map(function (programId) {
          return new PublicKey(programId).toBuffer();
        })
      };
      var signData = Buffer.alloc(2048);
      var length = signDataLayout.encode(transaction, signData);
      instructionBuffer.copy(signData, length);
      signData = signData.slice(0, length + instructionBuffer.length);
      return signData;
    }
    /**
     * Sign the Transaction with the specified accounts.  Multiple signatures may
     * be applied to a Transaction. The first signature is considered "primary"
     * and is used when testing for Transaction confirmation.
     *
     * Transaction fields should not be modified after the first call to `sign`,
     * as doing so may invalidate the signature and cause the Transaction to be
     * rejected.
     *
     * The Transaction must be assigned a valid `recentBlockhash` before invoking this method
     */

  }, {
    key: "sign",
    value: function sign() {
      this.signPartial.apply(this, arguments);
    }
    /**
     * Partially sign a Transaction with the specified accounts.  The `Account`
     * inputs will be used to sign the Transaction immediately, while any
     * `PublicKey` inputs will be referenced in the signed Transaction but need to
     * be filled in later by calling `addSigner()` with the matching `Account`.
     *
     * All the caveats from the `sign` method apply to `signPartial`
     */

  }, {
    key: "signPartial",
    value: function signPartial() {
      for (var _len2 = arguments.length, partialSigners = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
        partialSigners[_key2] = arguments[_key2];
      }

      if (partialSigners.length === 0) {
        throw new Error('No signers');
      }

      var signatures = partialSigners.map(function (accountOrPublicKey) {
        var publicKey = accountOrPublicKey instanceof Account ? accountOrPublicKey.publicKey : accountOrPublicKey;
        return {
          signature: null,
          publicKey: publicKey
        };
      });
      this.signatures = signatures;

      var signData = this._getSignData();

      partialSigners.forEach(function (accountOrPublicKey, index) {
        if (accountOrPublicKey instanceof PublicKey) {
          return;
        }

        var signature = nacl.sign.detached(signData, accountOrPublicKey.secretKey);
        assert(signature.length === 64);
        signatures[index].signature = Buffer.from(signature);
      });
    }
    /**
     * Fill in a signature for a partially signed Transaction.  The `signer` must
     * be the corresponding `Account` for a `PublicKey` that was previously provided to
     * `signPartial`
     */

  }, {
    key: "addSigner",
    value: function addSigner(signer) {
      var index = this.signatures.findIndex(function (sigpair) {
        return signer.publicKey.equals(sigpair.publicKey);
      });

      if (index < 0) {
        throw new Error("Unknown signer: ".concat(signer.publicKey.toString()));
      }

      var signData = this._getSignData();

      var signature = nacl.sign.detached(signData, signer.secretKey);
      assert(signature.length === 64);
      this.signatures[index].signature = Buffer.from(signature);
    }
    /**
     * Serialize the Transaction in the wire format.
     *
     * The Transaction must have a valid `signature` before invoking this method
     */

  }, {
    key: "serialize",
    value: function serialize() {
      var signatures = this.signatures;

      if (!signatures) {
        throw new Error('Transaction has not been signed');
      }

      var signData = this._getSignData();

      var signatureCount = [];
      encodeLength(signatureCount, signatures.length);
      var transactionLength = signatureCount.length + signatures.length * 64 + signData.length;
      var wireTransaction = Buffer.alloc(transactionLength);
      assert(signatures.length < 256);
      Buffer.from(signatureCount).copy(wireTransaction, 0);
      signatures.forEach(function (_ref2, index) {
        var signature = _ref2.signature;
        assert(signature !== null, "null signature");
        assert(signature.length === 64, "signature has invalid length");
        Buffer.from(signature).copy(wireTransaction, signatureCount.length + index * 64);
      });
      signData.copy(wireTransaction, signatureCount.length + signatures.length * 64);
      assert(wireTransaction.length <= PACKET_DATA_SIZE, "Transaction too large: ".concat(wireTransaction.length, " > ").concat(PACKET_DATA_SIZE));
      return wireTransaction;
    }
    /**
     * Deprecated method
     * @private
     */

  }, {
    key: "keys",
    get: function get() {
      assert(this.instructions.length === 1);
      return this.instructions[0].keys.map(function (keyObj) {
        return keyObj.pubkey;
      });
    }
    /**
     * Deprecated method
     * @private
     */

  }, {
    key: "programId",
    get: function get() {
      assert(this.instructions.length === 1);
      return this.instructions[0].programId;
    }
    /**
     * Deprecated method
     * @private
     */

  }, {
    key: "data",
    get: function get() {
      assert(this.instructions.length === 1);
      return this.instructions[0].data;
    }
    /**
     * Parse a wire transaction into a Transaction object.
     */

  }], [{
    key: "from",
    value: function from(buffer) {
      var PUBKEY_LENGTH = 32;
      var SIGNATURE_LENGTH = 64;
      var transaction = new Transaction(); // Slice up wire data

      var byteArray = _toConsumableArray(buffer);

      var signatureCount = decodeLength(byteArray);
      var signatures = [];

      for (var i = 0; i < signatureCount; i++) {
        var signature = byteArray.slice(0, SIGNATURE_LENGTH);
        byteArray = byteArray.slice(SIGNATURE_LENGTH);
        signatures.push(signature);
      }

      byteArray = byteArray.slice(1); // Skip numRequiredSignatures byte

      var accountCount = decodeLength(byteArray);
      var accounts = [];

      for (var _i = 0; _i < accountCount; _i++) {
        var account = byteArray.slice(0, PUBKEY_LENGTH);
        byteArray = byteArray.slice(PUBKEY_LENGTH);
        accounts.push(account);
      }

      var recentBlockhash = byteArray.slice(0, PUBKEY_LENGTH);
      byteArray = byteArray.slice(PUBKEY_LENGTH);
      var programIdCount = decodeLength(byteArray);
      var programs = [];

      for (var _i2 = 0; _i2 < programIdCount; _i2++) {
        var program = byteArray.slice(0, PUBKEY_LENGTH);
        byteArray = byteArray.slice(PUBKEY_LENGTH);
        programs.push(program);
      }

      var instructionCount = decodeLength(byteArray);
      var instructions = [];

      for (var _i3 = 0; _i3 < instructionCount; _i3++) {
        var instruction = {};
        instruction.programIndex = byteArray.shift();
        var accountIndexCount = decodeLength(byteArray);
        instruction.accountIndex = byteArray.slice(0, accountIndexCount);
        byteArray = byteArray.slice(accountIndexCount);
        var dataLength = decodeLength(byteArray);
        instruction.data = byteArray.slice(0, dataLength);
        byteArray = byteArray.slice(dataLength);
        instructions.push(instruction);
      } // Populate Transaction object


      transaction.recentBlockhash = new PublicKey(recentBlockhash).toBase58();

      for (var _i4 = 0; _i4 < signatureCount; _i4++) {
        var sigPubkeyPair = {
          signature: Buffer.from(signatures[_i4]),
          publicKey: new PublicKey(accounts[_i4])
        };
        transaction.signatures.push(sigPubkeyPair);
      }

      for (var _i5 = 0; _i5 < instructionCount; _i5++) {
        var instructionData = {
          keys: [],
          programId: new PublicKey(programs[instructions[_i5].programIndex]),
          data: Buffer.from(instructions[_i5].data)
        };

        var _loop = function _loop(j) {
          var pubkey = new PublicKey(accounts[instructions[_i5].accountIndex[j]]);
          instructionData.keys.push({
            pubkey: pubkey,
            isSigner: transaction.signatures.some(function (keyObj) {
              return keyObj.publicKey.toString() === pubkey.toString();
            })
          });
        };

        for (var j = 0; j < instructions[_i5].accountIndex.length; j++) {
          _loop(j);
        }

        var _instruction = new TransactionInstruction(instructionData);

        transaction.instructions.push(_instruction);
      }

      return transaction;
    }
  }]);

  return Transaction;
}();

// zzz
function sleep(ms) {
  return new Promise(function (resolve) {
    return setTimeout(resolve, ms);
  });
}

function createRpcRequest(url) {
  var server = jayson(
  /*#__PURE__*/
  function () {
    var _ref = _asyncToGenerator(
    /*#__PURE__*/
    _regeneratorRuntime.mark(function _callee(request, callback) {
      var options, res, text;
      return _regeneratorRuntime.wrap(function _callee$(_context) {
        while (1) {
          switch (_context.prev = _context.next) {
            case 0:
              options = {
                method: 'POST',
                body: request,
                headers: {
                  'Content-Type': 'application/json'
                }
              };
              _context.prev = 1;
              _context.next = 4;
              return fetch(url, options);

            case 4:
              res = _context.sent;
              _context.next = 7;
              return res.text();

            case 7:
              text = _context.sent;
              callback(null, text);
              _context.next = 14;
              break;

            case 11:
              _context.prev = 11;
              _context.t0 = _context["catch"](1);
              callback(_context.t0);

            case 14:
            case "end":
              return _context.stop();
          }
        }
      }, _callee, null, [[1, 11]]);
    }));

    return function (_x, _x2) {
      return _ref.apply(this, arguments);
    };
  }());
  return function (method, args) {
    return new Promise(function (resolve, reject) {
      server.request(method, args, function (err, response) {
        if (err) {
          reject(err);
          return;
        }

        resolve(response);
      });
    });
  };
}
/**
 * Expected JSON RPC response for the "getBalance" message
 */


var GetBalanceRpcResult = superstruct.struct({
  jsonrpc: superstruct.struct.literal('2.0'),
  id: 'string',
  error: 'any?',
  result: 'number?'
});
/**
 * @private
 */

function jsonRpcResult(resultDescription) {
  var jsonRpcVersion = superstruct.struct.literal('2.0');
  return superstruct.struct.union([superstruct.struct({
    jsonrpc: jsonRpcVersion,
    id: 'string',
    error: 'any'
  }), superstruct.struct({
    jsonrpc: jsonRpcVersion,
    id: 'string',
    error: 'null?',
    result: resultDescription
  })]);
}
/**
 * @private
 */


var AccountInfoResult = superstruct.struct({
  executable: 'boolean',
  owner: 'array',
  lamports: 'number',
  data: 'array'
});
/**
 * Expected JSON RPC response for the "getAccountInfo" message
 */

var GetAccountInfoRpcResult = jsonRpcResult(AccountInfoResult);
/***
 * Expected JSON RPC response for the "accountNotification" message
 */

var AccountNotificationResult = superstruct.struct({
  subscription: 'number',
  result: AccountInfoResult
});
/**
 * @private
 */

var ProgramAccountInfoResult = superstruct.struct(['string', AccountInfoResult]);
/***
 * Expected JSON RPC response for the "programNotification" message
 */

var ProgramAccountNotificationResult = superstruct.struct({
  subscription: 'number',
  result: ProgramAccountInfoResult
});
/**
 * Expected JSON RPC response for the "confirmTransaction" message
 */

var ConfirmTransactionRpcResult = jsonRpcResult('boolean');
/**
 * Expected JSON RPC response for the "getSignatureStatus" message
 */

var GetSignatureStatusRpcResult = jsonRpcResult(superstruct.struct.union(['null', superstruct.struct.union([superstruct.struct({
  Ok: 'null'
}), superstruct.struct({
  Err: 'object'
})])]));
/**
 * Expected JSON RPC response for the "getTransactionCount" message
 */

var GetTransactionCountRpcResult = jsonRpcResult('number');
/**
 * Expected JSON RPC response for the "getRecentBlockhash" message
 */

var GetRecentBlockhash = jsonRpcResult('string');
/**
 * Expected JSON RPC response for the "requestAirdrop" message
 */

var RequestAirdropRpcResult = jsonRpcResult('string');
/**
 * Expected JSON RPC response for the "sendTransaction" message
 */

var SendTransactionRpcResult = jsonRpcResult('string');
/**
 * Information describing an account
 *
 * @typedef {Object} AccountInfo
 * @property {number} lamports Number of lamports assigned to the account
 * @property {PublicKey} owner Identifier of the program that owns the account
 * @property {?Buffer} data Optional data assigned to the account
 * @property {boolean} executable `true` if this account's data contains a loaded program
 */

/**
 * A connection to a fullnode JSON RPC endpoint
 */
var Connection =
/*#__PURE__*/
function () {
  /**
   * Establish a JSON RPC connection
   *
   * @param endpoint URL to the fullnode JSON RPC endpoint
   */
  function Connection(endpoint) {
    _classCallCheck(this, Connection);

    _defineProperty(this, "_rpcRequest", void 0);

    _defineProperty(this, "_rpcWebSocket", void 0);

    _defineProperty(this, "_rpcWebSocketConnected", false);

    _defineProperty(this, "_blockhashInfo", void 0);

    _defineProperty(this, "_disableBlockhashCaching", false);

    _defineProperty(this, "_accountChangeSubscriptions", {});

    _defineProperty(this, "_accountChangeSubscriptionCounter", 0);

    _defineProperty(this, "_programAccountChangeSubscriptions", {});

    _defineProperty(this, "_programAccountChangeSubscriptionCounter", 0);

    var url$1 = url.parse(endpoint);
    this._rpcRequest = createRpcRequest(url$1.href);
    this._blockhashInfo = {
      recentBlockhash: null,
      seconds: -1,
      transactionSignatures: []
    };
    url$1.protocol = url$1.protocol === 'https:' ? 'wss:' : 'ws:';
    url$1.host = '';
    url$1.port = String(Number(url$1.port) + 1);

    if (url$1.port === '1') {
      url$1.port = url$1.protocol === 'wss:' ? '8901' : '8900';
    }

    this._rpcWebSocket = new rpcWebsockets.Client(url.format(url$1), {
      autoconnect: false,
      max_reconnects: Infinity
    });

    this._rpcWebSocket.on('open', this._wsOnOpen.bind(this));

    this._rpcWebSocket.on('error', this._wsOnError.bind(this));

    this._rpcWebSocket.on('close', this._wsOnClose.bind(this));

    this._rpcWebSocket.on('accountNotification', this._wsOnAccountNotification.bind(this));

    this._rpcWebSocket.on('programNotification', this._wsOnProgramAccountNotification.bind(this));
  }
  /**
   * Fetch the balance for the specified public key
   */


  _createClass(Connection, [{
    key: "getBalance",
    value: function () {
      var _getBalance = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee2(publicKey) {
        var unsafeRes, res;
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _context2.next = 2;
                return this._rpcRequest('getBalance', [publicKey.toBase58()]);

              case 2:
                unsafeRes = _context2.sent;
                res = GetBalanceRpcResult(unsafeRes);

                if (!res.error) {
                  _context2.next = 6;
                  break;
                }

                throw new Error(res.error.message);

              case 6:
                assert(typeof res.result !== 'undefined');
                return _context2.abrupt("return", res.result);

              case 8:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function getBalance(_x3) {
        return _getBalance.apply(this, arguments);
      }

      return getBalance;
    }()
    /**
     * Fetch all the account info for the specified public key
     */

  }, {
    key: "getAccountInfo",
    value: function () {
      var _getAccountInfo = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee3(publicKey) {
        var unsafeRes, res, result;
        return _regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                _context3.next = 2;
                return this._rpcRequest('getAccountInfo', [publicKey.toBase58()]);

              case 2:
                unsafeRes = _context3.sent;
                res = GetAccountInfoRpcResult(unsafeRes);

                if (!res.error) {
                  _context3.next = 6;
                  break;
                }

                throw new Error(res.error.message);

              case 6:
                result = res.result;
                assert(typeof result !== 'undefined');
                return _context3.abrupt("return", {
                  executable: result.executable,
                  owner: new PublicKey(result.owner),
                  lamports: result.lamports,
                  data: Buffer.from(result.data)
                });

              case 9:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function getAccountInfo(_x4) {
        return _getAccountInfo.apply(this, arguments);
      }

      return getAccountInfo;
    }()
    /**
     * Confirm the transaction identified by the specified signature
     */

  }, {
    key: "confirmTransaction",
    value: function () {
      var _confirmTransaction = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee4(signature) {
        var unsafeRes, res;
        return _regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                _context4.next = 2;
                return this._rpcRequest('confirmTransaction', [signature]);

              case 2:
                unsafeRes = _context4.sent;
                res = ConfirmTransactionRpcResult(unsafeRes);

                if (!res.error) {
                  _context4.next = 6;
                  break;
                }

                throw new Error(res.error.message);

              case 6:
                assert(typeof res.result !== 'undefined');
                return _context4.abrupt("return", res.result);

              case 8:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function confirmTransaction(_x5) {
        return _confirmTransaction.apply(this, arguments);
      }

      return confirmTransaction;
    }()
    /**
     * Fetch the current transaction count of the cluster
     */

  }, {
    key: "getSignatureStatus",
    value: function () {
      var _getSignatureStatus = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee5(signature) {
        var unsafeRes, res;
        return _regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                _context5.next = 2;
                return this._rpcRequest('getSignatureStatus', [signature]);

              case 2:
                unsafeRes = _context5.sent;
                res = GetSignatureStatusRpcResult(unsafeRes);

                if (!res.error) {
                  _context5.next = 6;
                  break;
                }

                throw new Error(res.error.message);

              case 6:
                assert(typeof res.result !== 'undefined');
                return _context5.abrupt("return", res.result);

              case 8:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5, this);
      }));

      function getSignatureStatus(_x6) {
        return _getSignatureStatus.apply(this, arguments);
      }

      return getSignatureStatus;
    }()
    /**
     * Fetch the current transaction count of the cluster
     */

  }, {
    key: "getTransactionCount",
    value: function () {
      var _getTransactionCount = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee6() {
        var unsafeRes, res;
        return _regeneratorRuntime.wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                _context6.next = 2;
                return this._rpcRequest('getTransactionCount', []);

              case 2:
                unsafeRes = _context6.sent;
                res = GetTransactionCountRpcResult(unsafeRes);

                if (!res.error) {
                  _context6.next = 6;
                  break;
                }

                throw new Error(res.error.message);

              case 6:
                assert(typeof res.result !== 'undefined');
                return _context6.abrupt("return", Number(res.result));

              case 8:
              case "end":
                return _context6.stop();
            }
          }
        }, _callee6, this);
      }));

      function getTransactionCount() {
        return _getTransactionCount.apply(this, arguments);
      }

      return getTransactionCount;
    }()
    /**
     * Fetch a recent blockhash from the cluster
     */

  }, {
    key: "getRecentBlockhash",
    value: function () {
      var _getRecentBlockhash = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee7() {
        var unsafeRes, res;
        return _regeneratorRuntime.wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                _context7.next = 2;
                return this._rpcRequest('getRecentBlockhash', []);

              case 2:
                unsafeRes = _context7.sent;
                res = GetRecentBlockhash(unsafeRes);

                if (!res.error) {
                  _context7.next = 6;
                  break;
                }

                throw new Error(res.error.message);

              case 6:
                assert(typeof res.result !== 'undefined');
                return _context7.abrupt("return", res.result);

              case 8:
              case "end":
                return _context7.stop();
            }
          }
        }, _callee7, this);
      }));

      function getRecentBlockhash() {
        return _getRecentBlockhash.apply(this, arguments);
      }

      return getRecentBlockhash;
    }()
    /**
     * Request an allocation of lamports to the specified account
     */

  }, {
    key: "requestAirdrop",
    value: function () {
      var _requestAirdrop = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee8(to, amount) {
        var unsafeRes, res;
        return _regeneratorRuntime.wrap(function _callee8$(_context8) {
          while (1) {
            switch (_context8.prev = _context8.next) {
              case 0:
                _context8.next = 2;
                return this._rpcRequest('requestAirdrop', [to.toBase58(), amount]);

              case 2:
                unsafeRes = _context8.sent;
                res = RequestAirdropRpcResult(unsafeRes);

                if (!res.error) {
                  _context8.next = 6;
                  break;
                }

                throw new Error(res.error.message);

              case 6:
                assert(typeof res.result !== 'undefined');
                return _context8.abrupt("return", res.result);

              case 8:
              case "end":
                return _context8.stop();
            }
          }
        }, _callee8, this);
      }));

      function requestAirdrop(_x7, _x8) {
        return _requestAirdrop.apply(this, arguments);
      }

      return requestAirdrop;
    }()
    /**
     * Sign and send a transaction
     */

  }, {
    key: "sendTransaction",
    value: function () {
      var _sendTransaction = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee9(transaction) {
        var _len,
            signers,
            _key,
            seconds,
            signature,
            attempts,
            startTime,
            recentBlockhash,
            wireTransaction,
            _args9 = arguments;

        return _regeneratorRuntime.wrap(function _callee9$(_context9) {
          while (1) {
            switch (_context9.prev = _context9.next) {
              case 0:
                for (_len = _args9.length, signers = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
                  signers[_key - 1] = _args9[_key];
                }

              case 1:
                // Attempt to use a recent blockhash for up to 30 seconds
                seconds = new Date().getSeconds();

                if (!(this._blockhashInfo.recentBlockhash != null && this._blockhashInfo.seconds < seconds + 30)) {
                  _context9.next = 12;
                  break;
                }

                transaction.recentBlockhash = this._blockhashInfo.recentBlockhash;
                transaction.sign.apply(transaction, signers);

                if (transaction.signature) {
                  _context9.next = 7;
                  break;
                }

                throw new Error('!signature');

              case 7:
                // If the signature of this transaction has not been seen before with the
                // current recentBlockhash, all done.
                signature = transaction.signature.toString();

                if (this._blockhashInfo.transactionSignatures.includes(signature)) {
                  _context9.next = 12;
                  break;
                }

                this._blockhashInfo.transactionSignatures.push(signature);

                if (this._disableBlockhashCaching) {
                  this._blockhashInfo.seconds = -1;
                }

                return _context9.abrupt("break", 29);

              case 12:
                // Fetch a new blockhash
                attempts = 0;
                startTime = Date.now();

              case 14:
                _context9.next = 16;
                return this.getRecentBlockhash();

              case 16:
                recentBlockhash = _context9.sent;

                if (!(this._blockhashInfo.recentBlockhash != recentBlockhash)) {
                  _context9.next = 20;
                  break;
                }

                this._blockhashInfo = {
                  recentBlockhash: recentBlockhash,
                  seconds: new Date().getSeconds(),
                  transactionSignatures: []
                };
                return _context9.abrupt("break", 27);

              case 20:
                if (!(attempts === 16)) {
                  _context9.next = 22;
                  break;
                }

                throw new Error("Unable to obtain a new blockhash after ".concat(Date.now() - startTime, "ms"));

              case 22:
                _context9.next = 24;
                return sleep(500 * DEFAULT_TICKS_PER_SLOT / NUM_TICKS_PER_SECOND);

              case 24:
                ++attempts;

              case 25:
                _context9.next = 14;
                break;

              case 27:
                _context9.next = 1;
                break;

              case 29:
                wireTransaction = transaction.serialize();
                _context9.next = 32;
                return this.sendRawTransaction(wireTransaction);

              case 32:
                return _context9.abrupt("return", _context9.sent);

              case 33:
              case "end":
                return _context9.stop();
            }
          }
        }, _callee9, this);
      }));

      function sendTransaction(_x9) {
        return _sendTransaction.apply(this, arguments);
      }

      return sendTransaction;
    }()
    /**
     * Send a transaction that has already been signed and serialized into the
     * wire format
     */

  }, {
    key: "sendRawTransaction",
    value: function () {
      var _sendRawTransaction = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee10(rawTransaction) {
        var unsafeRes, res;
        return _regeneratorRuntime.wrap(function _callee10$(_context10) {
          while (1) {
            switch (_context10.prev = _context10.next) {
              case 0:
                _context10.next = 2;
                return this._rpcRequest('sendTransaction', [_toConsumableArray(rawTransaction)]);

              case 2:
                unsafeRes = _context10.sent;
                res = SendTransactionRpcResult(unsafeRes);

                if (!res.error) {
                  _context10.next = 6;
                  break;
                }

                throw new Error(res.error.message);

              case 6:
                assert(typeof res.result !== 'undefined');
                assert(res.result);
                return _context10.abrupt("return", res.result);

              case 9:
              case "end":
                return _context10.stop();
            }
          }
        }, _callee10, this);
      }));

      function sendRawTransaction(_x10) {
        return _sendRawTransaction.apply(this, arguments);
      }

      return sendRawTransaction;
    }()
    /**
     * @private
     */

  }, {
    key: "_wsOnOpen",
    value: function _wsOnOpen() {
      this._rpcWebSocketConnected = true;

      this._updateSubscriptions();
    }
    /**
     * @private
     */

  }, {
    key: "_wsOnError",
    value: function _wsOnError(err) {
      console.log('ws error:', err.message);
    }
    /**
     * @private
     */

  }, {
    key: "_wsOnClose",
    value: function _wsOnClose(code, message) {
      // 1000 means _rpcWebSocket.close() was called explicitly
      if (code !== 1000) {
        console.log('ws close:', code, message);
      }

      this._rpcWebSocketConnected = false;
    }
    /**
     * @private
     */

  }, {
    key: "_updateSubscriptions",
    value: function () {
      var _updateSubscriptions2 = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee11() {
        var accountKeys, programKeys, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, id, _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, _id, _iteratorNormalCompletion3, _didIteratorError3, _iteratorError3, _iterator3, _step3, _id2, _this$_accountChangeS, subscriptionId, publicKey, _iteratorNormalCompletion4, _didIteratorError4, _iteratorError4, _iterator4, _step4, _id3, _this$_programAccount, programId;

        return _regeneratorRuntime.wrap(function _callee11$(_context11) {
          while (1) {
            switch (_context11.prev = _context11.next) {
              case 0:
                accountKeys = Object.keys(this._accountChangeSubscriptions).map(Number);
                programKeys = Object.keys(this._programAccountChangeSubscriptions).map(Number);

                if (!(accountKeys.length === 0 && programKeys.length === 0)) {
                  _context11.next = 5;
                  break;
                }

                this._rpcWebSocket.close();

                return _context11.abrupt("return");

              case 5:
                if (this._rpcWebSocketConnected) {
                  _context11.next = 46;
                  break;
                }

                _iteratorNormalCompletion = true;
                _didIteratorError = false;
                _iteratorError = undefined;
                _context11.prev = 9;

                for (_iterator = accountKeys[Symbol.iterator](); !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                  id = _step.value;
                  this._accountChangeSubscriptions[id].subscriptionId = null;
                }

                _context11.next = 17;
                break;

              case 13:
                _context11.prev = 13;
                _context11.t0 = _context11["catch"](9);
                _didIteratorError = true;
                _iteratorError = _context11.t0;

              case 17:
                _context11.prev = 17;
                _context11.prev = 18;

                if (!_iteratorNormalCompletion && _iterator.return != null) {
                  _iterator.return();
                }

              case 20:
                _context11.prev = 20;

                if (!_didIteratorError) {
                  _context11.next = 23;
                  break;
                }

                throw _iteratorError;

              case 23:
                return _context11.finish(20);

              case 24:
                return _context11.finish(17);

              case 25:
                _iteratorNormalCompletion2 = true;
                _didIteratorError2 = false;
                _iteratorError2 = undefined;
                _context11.prev = 28;

                for (_iterator2 = programKeys[Symbol.iterator](); !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                  _id = _step2.value;
                  this._programAccountChangeSubscriptions[_id].subscriptionId = null;
                }

                _context11.next = 36;
                break;

              case 32:
                _context11.prev = 32;
                _context11.t1 = _context11["catch"](28);
                _didIteratorError2 = true;
                _iteratorError2 = _context11.t1;

              case 36:
                _context11.prev = 36;
                _context11.prev = 37;

                if (!_iteratorNormalCompletion2 && _iterator2.return != null) {
                  _iterator2.return();
                }

              case 39:
                _context11.prev = 39;

                if (!_didIteratorError2) {
                  _context11.next = 42;
                  break;
                }

                throw _iteratorError2;

              case 42:
                return _context11.finish(39);

              case 43:
                return _context11.finish(36);

              case 44:
                this._rpcWebSocket.connect();

                return _context11.abrupt("return");

              case 46:
                _iteratorNormalCompletion3 = true;
                _didIteratorError3 = false;
                _iteratorError3 = undefined;
                _context11.prev = 49;
                _iterator3 = accountKeys[Symbol.iterator]();

              case 51:
                if (_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done) {
                  _context11.next = 67;
                  break;
                }

                _id2 = _step3.value;
                _this$_accountChangeS = this._accountChangeSubscriptions[_id2], subscriptionId = _this$_accountChangeS.subscriptionId, publicKey = _this$_accountChangeS.publicKey;

                if (!(subscriptionId === null)) {
                  _context11.next = 64;
                  break;
                }

                _context11.prev = 55;
                _context11.next = 58;
                return this._rpcWebSocket.call('accountSubscribe', [publicKey]);

              case 58:
                this._accountChangeSubscriptions[_id2].subscriptionId = _context11.sent;
                _context11.next = 64;
                break;

              case 61:
                _context11.prev = 61;
                _context11.t2 = _context11["catch"](55);
                console.log("accountSubscribe error for ".concat(publicKey, ": ").concat(_context11.t2.message));

              case 64:
                _iteratorNormalCompletion3 = true;
                _context11.next = 51;
                break;

              case 67:
                _context11.next = 73;
                break;

              case 69:
                _context11.prev = 69;
                _context11.t3 = _context11["catch"](49);
                _didIteratorError3 = true;
                _iteratorError3 = _context11.t3;

              case 73:
                _context11.prev = 73;
                _context11.prev = 74;

                if (!_iteratorNormalCompletion3 && _iterator3.return != null) {
                  _iterator3.return();
                }

              case 76:
                _context11.prev = 76;

                if (!_didIteratorError3) {
                  _context11.next = 79;
                  break;
                }

                throw _iteratorError3;

              case 79:
                return _context11.finish(76);

              case 80:
                return _context11.finish(73);

              case 81:
                _iteratorNormalCompletion4 = true;
                _didIteratorError4 = false;
                _iteratorError4 = undefined;
                _context11.prev = 84;
                _iterator4 = programKeys[Symbol.iterator]();

              case 86:
                if (_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done) {
                  _context11.next = 103;
                  break;
                }

                _id3 = _step4.value;
                _this$_programAccount = this._programAccountChangeSubscriptions[_id3], subscriptionId = _this$_programAccount.subscriptionId, programId = _this$_programAccount.programId;
                console.log('program-id: ' + programId);

                if (!(subscriptionId === null)) {
                  _context11.next = 100;
                  break;
                }

                _context11.prev = 91;
                _context11.next = 94;
                return this._rpcWebSocket.call('programSubscribe', [programId]);

              case 94:
                this._programAccountChangeSubscriptions[_id3].subscriptionId = _context11.sent;
                _context11.next = 100;
                break;

              case 97:
                _context11.prev = 97;
                _context11.t4 = _context11["catch"](91);
                console.log("programSubscribe error for ".concat(programId, ": ").concat(_context11.t4.message));

              case 100:
                _iteratorNormalCompletion4 = true;
                _context11.next = 86;
                break;

              case 103:
                _context11.next = 109;
                break;

              case 105:
                _context11.prev = 105;
                _context11.t5 = _context11["catch"](84);
                _didIteratorError4 = true;
                _iteratorError4 = _context11.t5;

              case 109:
                _context11.prev = 109;
                _context11.prev = 110;

                if (!_iteratorNormalCompletion4 && _iterator4.return != null) {
                  _iterator4.return();
                }

              case 112:
                _context11.prev = 112;

                if (!_didIteratorError4) {
                  _context11.next = 115;
                  break;
                }

                throw _iteratorError4;

              case 115:
                return _context11.finish(112);

              case 116:
                return _context11.finish(109);

              case 117:
              case "end":
                return _context11.stop();
            }
          }
        }, _callee11, this, [[9, 13, 17, 25], [18,, 20, 24], [28, 32, 36, 44], [37,, 39, 43], [49, 69, 73, 81], [55, 61], [74,, 76, 80], [84, 105, 109, 117], [91, 97], [110,, 112, 116]]);
      }));

      function _updateSubscriptions() {
        return _updateSubscriptions2.apply(this, arguments);
      }

      return _updateSubscriptions;
    }()
    /**
     * @private
     */

  }, {
    key: "_wsOnAccountNotification",
    value: function _wsOnAccountNotification(notification) {
      var res = AccountNotificationResult(notification);

      if (res.error) {
        throw new Error(res.error.message);
      }

      var keys = Object.keys(this._accountChangeSubscriptions).map(Number);
      var _iteratorNormalCompletion5 = true;
      var _didIteratorError5 = false;
      var _iteratorError5 = undefined;

      try {
        for (var _iterator5 = keys[Symbol.iterator](), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
          var id = _step5.value;
          var sub = this._accountChangeSubscriptions[id];

          if (sub.subscriptionId === res.subscription) {
            var result = res.result;
            assert(typeof result !== 'undefined');
            sub.callback({
              executable: result.executable,
              owner: new PublicKey(result.owner),
              lamports: result.lamports,
              data: Buffer.from(result.data)
            });
            return true;
          }
        }
      } catch (err) {
        _didIteratorError5 = true;
        _iteratorError5 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion5 && _iterator5.return != null) {
            _iterator5.return();
          }
        } finally {
          if (_didIteratorError5) {
            throw _iteratorError5;
          }
        }
      }
    }
    /**
     * Register a callback to be invoked whenever the specified account changes
     *
     * @param publickey Public key of the account to monitor
     * @param callback Function to invoke whenever the account is changed
     * @return subscription id
     */

  }, {
    key: "onAccountChange",
    value: function onAccountChange(publicKey, callback) {
      var id = ++this._accountChangeSubscriptionCounter;
      this._accountChangeSubscriptions[id] = {
        publicKey: publicKey.toBase58(),
        callback: callback,
        subscriptionId: null
      };

      this._updateSubscriptions();

      return id;
    }
    /**
     * Deregister an account notification callback
     *
     * @param id subscription id to deregister
     */

  }, {
    key: "removeAccountChangeListener",
    value: function () {
      var _removeAccountChangeListener = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee12(id) {
        var subscriptionId;
        return _regeneratorRuntime.wrap(function _callee12$(_context12) {
          while (1) {
            switch (_context12.prev = _context12.next) {
              case 0:
                if (!this._accountChangeSubscriptions[id]) {
                  _context12.next = 15;
                  break;
                }

                subscriptionId = this._accountChangeSubscriptions[id].subscriptionId;
                delete this._accountChangeSubscriptions[id];

                if (!(subscriptionId !== null)) {
                  _context12.next = 12;
                  break;
                }

                _context12.prev = 4;
                _context12.next = 7;
                return this._rpcWebSocket.call('accountUnsubscribe', [subscriptionId]);

              case 7:
                _context12.next = 12;
                break;

              case 9:
                _context12.prev = 9;
                _context12.t0 = _context12["catch"](4);
                console.log('accountUnsubscribe error:', _context12.t0.message);

              case 12:
                this._updateSubscriptions();

                _context12.next = 16;
                break;

              case 15:
                throw new Error("Unknown account change id: ".concat(id));

              case 16:
              case "end":
                return _context12.stop();
            }
          }
        }, _callee12, this, [[4, 9]]);
      }));

      function removeAccountChangeListener(_x11) {
        return _removeAccountChangeListener.apply(this, arguments);
      }

      return removeAccountChangeListener;
    }()
    /**
     * @private
     */

  }, {
    key: "_wsOnProgramAccountNotification",
    value: function _wsOnProgramAccountNotification(notification) {
      var res = ProgramAccountNotificationResult(notification);

      if (res.error) {
        throw new Error(res.error.message);
      }

      var keys = Object.keys(this._programAccountChangeSubscriptions).map(Number);
      var _iteratorNormalCompletion6 = true;
      var _didIteratorError6 = false;
      var _iteratorError6 = undefined;

      try {
        for (var _iterator6 = keys[Symbol.iterator](), _step6; !(_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done); _iteratorNormalCompletion6 = true) {
          var id = _step6.value;
          var sub = this._programAccountChangeSubscriptions[id];

          if (sub.subscriptionId === res.subscription) {
            var result = res.result;
            assert(typeof result !== 'undefined');
            sub.callback({
              accountId: result[0],
              accountInfo: {
                executable: result[1].executable,
                owner: new PublicKey(result[1].owner),
                lamports: result[1].lamports,
                data: Buffer.from(result[1].data)
              }
            });
            return true;
          }
        }
      } catch (err) {
        _didIteratorError6 = true;
        _iteratorError6 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion6 && _iterator6.return != null) {
            _iterator6.return();
          }
        } finally {
          if (_didIteratorError6) {
            throw _iteratorError6;
          }
        }
      }
    }
    /**
     * Register a callback to be invoked whenever accounts owned by the
     * specified program change
     *
     * @param programId Public key of the program to monitor
     * @param callback Function to invoke whenever the account is changed
     * @return subscription id
     */

  }, {
    key: "onProgramAccountChange",
    value: function onProgramAccountChange(programId, callback) {
      var id = ++this._programAccountChangeSubscriptionCounter;
      this._programAccountChangeSubscriptions[id] = {
        programId: programId.toBase58(),
        callback: callback,
        subscriptionId: null
      };

      this._updateSubscriptions();

      return id;
    }
    /**
     * Deregister an account notification callback
     *
     * @param id subscription id to deregister
     */

  }, {
    key: "removeProgramAccountChangeListener",
    value: function () {
      var _removeProgramAccountChangeListener = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee13(id) {
        var subscriptionId;
        return _regeneratorRuntime.wrap(function _callee13$(_context13) {
          while (1) {
            switch (_context13.prev = _context13.next) {
              case 0:
                if (!this._programAccountChangeSubscriptions[id]) {
                  _context13.next = 15;
                  break;
                }

                subscriptionId = this._programAccountChangeSubscriptions[id].subscriptionId;
                delete this._programAccountChangeSubscriptions[id];

                if (!(subscriptionId !== null)) {
                  _context13.next = 12;
                  break;
                }

                _context13.prev = 4;
                _context13.next = 7;
                return this._rpcWebSocket.call('programUnsubscribe', [subscriptionId]);

              case 7:
                _context13.next = 12;
                break;

              case 9:
                _context13.prev = 9;
                _context13.t0 = _context13["catch"](4);
                console.log('programUnsubscribe error:', _context13.t0.message);

              case 12:
                this._updateSubscriptions();

                _context13.next = 16;
                break;

              case 15:
                throw new Error("Unknown account change id: ".concat(id));

              case 16:
              case "end":
                return _context13.stop();
            }
          }
        }, _callee13, this, [[4, 9]]);
      }));

      function removeProgramAccountChangeListener(_x12) {
        return _removeProgramAccountChangeListener.apply(this, arguments);
      }

      return removeProgramAccountChangeListener;
    }()
  }]);

  return Connection;
}();

/**
 * Sign, send and confirm a transaction
 */

function sendAndConfirmTransaction(_x, _x2) {
  return _sendAndConfirmTransaction.apply(this, arguments);
}

function _sendAndConfirmTransaction() {
  _sendAndConfirmTransaction = _asyncToGenerator(
  /*#__PURE__*/
  _regeneratorRuntime.mark(function _callee(connection, transaction) {
    var sendRetries,
        signature,
        _len,
        signers,
        _key,
        start,
        status,
        statusRetries,
        duration,
        _args = arguments;

    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            sendRetries = 10;

            for (_len = _args.length, signers = new Array(_len > 2 ? _len - 2 : 0), _key = 2; _key < _len; _key++) {
              signers[_key - 2] = _args[_key];
            }

          case 2:
            start = Date.now();
            _context.next = 5;
            return connection.sendTransaction.apply(connection, [transaction].concat(signers));

          case 5:
            signature = _context.sent;
            // Wait up to a couple slots for a confirmation
            status = null;
            statusRetries = 6;

          case 8:
            _context.next = 10;
            return connection.getSignatureStatus(signature);

          case 10:
            status = _context.sent;

            if (!status) {
              _context.next = 13;
              break;
            }

            return _context.abrupt("break", 19);

          case 13:
            if (!(--statusRetries <= 0)) {
              _context.next = 15;
              break;
            }

            return _context.abrupt("break", 19);

          case 15:
            _context.next = 17;
            return sleep(500 * DEFAULT_TICKS_PER_SLOT / NUM_TICKS_PER_SECOND);

          case 17:
            _context.next = 8;
            break;

          case 19:
            if (!(status && 'Ok' in status)) {
              _context.next = 21;
              break;
            }

            return _context.abrupt("break", 30);

          case 21:
            if (!(--sendRetries <= 0)) {
              _context.next = 24;
              break;
            }

            duration = (Date.now() - start) / 1000;
            throw new Error("Transaction '".concat(signature, "' was not confirmed in ").concat(duration.toFixed(2), " seconds (").concat(JSON.stringify(status), ")"));

          case 24:
            if (!(status && status.Err && !('AccountInUse' in status.Err))) {
              _context.next = 26;
              break;
            }

            throw new Error("Transaction ".concat(signature, " failed (").concat(JSON.stringify(status), ")"));

          case 26:
            _context.next = 28;
            return sleep(Math.random() * 100);

          case 28:
            _context.next = 2;
            break;

          case 30:
            assert(signature !== undefined);
            return _context.abrupt("return", signature);

          case 32:
          case "end":
            return _context.stop();
        }
      }
    }, _callee);
  }));
  return _sendAndConfirmTransaction.apply(this, arguments);
}

/**
 * Program loader interface
 */
var Loader =
/*#__PURE__*/
function () {
  _createClass(Loader, null, [{
    key: "chunkSize",

    /**
     * @private
     */

    /**
     * @private
     */

    /**
     * Amount of program data placed in each load Transaction
     */
    get: function get() {
      return 256;
    }
    /**
     * @param connection The connection to use
     * @param programId Public key that identifies the loader
     */

  }]);

  function Loader(connection, programId) {
    _classCallCheck(this, Loader);

    _defineProperty(this, "connection", void 0);

    _defineProperty(this, "programId", void 0);

    Object.assign(this, {
      connection: connection,
      programId: programId
    });
  }
  /**
   * Load program data
   *
   * @param program Account to load the program info
   * @param data Program data
   */


  _createClass(Loader, [{
    key: "load",
    value: function () {
      var _load = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee(program, data) {
        var dataLayout, chunkSize, offset, array, transactions, bytes, _data, transaction;

        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                dataLayout = BufferLayout.struct([BufferLayout.u32('instruction'), BufferLayout.u32('offset'), BufferLayout.u32('bytesLength'), BufferLayout.u32('bytesLengthPadding'), BufferLayout.seq(BufferLayout.u8('byte'), BufferLayout.offset(BufferLayout.u32(), -8), 'bytes')]);
                chunkSize = Loader.chunkSize;
                offset = 0;
                array = data;
                transactions = [];

              case 5:
                if (!(array.length > 0)) {
                  _context.next = 21;
                  break;
                }

                bytes = array.slice(0, chunkSize);
                _data = Buffer.alloc(chunkSize + 16);
                dataLayout.encode({
                  instruction: 0,
                  // Load instruction
                  offset: offset,
                  bytes: bytes
                }, _data);
                transaction = new Transaction().add({
                  keys: [{
                    pubkey: program.publicKey,
                    isSigner: true
                  }],
                  programId: this.programId,
                  data: _data
                });
                transactions.push(sendAndConfirmTransaction(this.connection, transaction, program)); // Delay ~1 tick between write transactions in an attempt to reduce AccountInUse errors
                // since all the write transactions modify the same program account

                _context.next = 13;
                return sleep(1000 / NUM_TICKS_PER_SECOND);

              case 13:
                if (!(transactions.length === 8)) {
                  _context.next = 17;
                  break;
                }

                _context.next = 16;
                return Promise.all(transactions);

              case 16:
                transactions = [];

              case 17:
                offset += chunkSize;
                array = array.slice(chunkSize);
                _context.next = 5;
                break;

              case 21:
                _context.next = 23;
                return Promise.all(transactions);

              case 23:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function load(_x, _x2) {
        return _load.apply(this, arguments);
      }

      return load;
    }()
    /**
     * Finalize an account loaded with program data for execution
     *
     * @param program `load()`ed Account
     */

  }, {
    key: "finalize",
    value: function () {
      var _finalize = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee2(program) {
        var dataLayout, data, transaction;
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                dataLayout = BufferLayout.struct([BufferLayout.u32('instruction')]);
                data = Buffer.alloc(dataLayout.span);
                dataLayout.encode({
                  instruction: 1 // Finalize instruction

                }, data);
                transaction = new Transaction().add({
                  keys: [{
                    pubkey: program.publicKey,
                    isSigner: true
                  }],
                  programId: this.programId,
                  data: data
                });
                _context2.next = 6;
                return sendAndConfirmTransaction(this.connection, transaction, program);

              case 6:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function finalize(_x3) {
        return _finalize.apply(this, arguments);
      }

      return finalize;
    }()
  }]);

  return Loader;
}();

/**
 * Factory class for transactions to interact with the System program
 */

var SystemProgram =
/*#__PURE__*/
function () {
  function SystemProgram() {
    _classCallCheck(this, SystemProgram);
  }

  _createClass(SystemProgram, null, [{
    key: "createAccount",

    /**
     * Generate a Transaction that creates a new account
     */
    value: function createAccount(from, newAccount, lamports, space, programId) {
      var dataLayout = BufferLayout.struct([BufferLayout.u32('instruction'), BufferLayout.ns64('lamports'), BufferLayout.ns64('space'), publicKey('programId')]);
      var data = Buffer.alloc(dataLayout.span);
      dataLayout.encode({
        instruction: 0,
        // Create Account instruction
        lamports: lamports,
        space: space,
        programId: programId.toBuffer()
      }, data);
      return new Transaction().add({
        keys: [{
          pubkey: from,
          isSigner: true
        }, {
          pubkey: newAccount,
          isSigner: false
        }],
        programId: SystemProgram.programId,
        data: data
      });
    }
    /**
     * Generate a Transaction that transfers lamports from one account to another
     */

  }, {
    key: "transfer",
    value: function transfer(from, to, amount) {
      var dataLayout = BufferLayout.struct([BufferLayout.u32('instruction'), BufferLayout.ns64('amount')]);
      var data = Buffer.alloc(dataLayout.span);
      dataLayout.encode({
        instruction: 2,
        // Move instruction
        amount: amount
      }, data);
      return new Transaction().add({
        keys: [{
          pubkey: from,
          isSigner: true
        }, {
          pubkey: to,
          isSigner: false
        }],
        programId: SystemProgram.programId,
        data: data
      });
    }
    /**
     * Generate a Transaction that assigns an account to a program
     */

  }, {
    key: "assign",
    value: function assign(from, programId) {
      var dataLayout = BufferLayout.struct([BufferLayout.u32('instruction'), publicKey('programId')]);
      var data = Buffer.alloc(dataLayout.span);
      dataLayout.encode({
        instruction: 1,
        // Assign instruction
        programId: programId.toBuffer()
      }, data);
      return new Transaction().add({
        keys: [{
          pubkey: from,
          isSigner: true
        }],
        programId: SystemProgram.programId,
        data: data
      });
    }
  }, {
    key: "programId",

    /**
     * Public key that identifies the System program
     */
    get: function get() {
      return new PublicKey('0x000000000000000000000000000000000000000000000000000000000000000');
    }
  }]);

  return SystemProgram;
}();

/**
 * Factory class for transactions to interact with a program loader
 */
var BpfLoader =
/*#__PURE__*/
function () {
  function BpfLoader() {
    _classCallCheck(this, BpfLoader);
  }

  _createClass(BpfLoader, null, [{
    key: "load",

    /**
     * Load a BPF program
     *
     * @param connection The connection to use
     * @param owner User account to load the program into
     * @param elfBytes The entire ELF containing the BPF program
     */
    value: function () {
      var _load = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee(connection, owner, elf) {
        var programAccount, transaction, loader;
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                programAccount = new Account();
                transaction = SystemProgram.createAccount(owner.publicKey, programAccount.publicKey, 1 + Math.ceil(elf.length / Loader.chunkSize) + 1, elf.length, BpfLoader.programId);
                _context.next = 4;
                return sendAndConfirmTransaction(connection, transaction, owner);

              case 4:
                loader = new Loader(connection, BpfLoader.programId);
                _context.next = 7;
                return loader.load(programAccount, elf);

              case 7:
                _context.next = 9;
                return loader.finalize(programAccount);

              case 9:
                return _context.abrupt("return", programAccount.publicKey);

              case 10:
              case "end":
                return _context.stop();
            }
          }
        }, _callee);
      }));

      function load(_x, _x2, _x3) {
        return _load.apply(this, arguments);
      }

      return load;
    }()
  }, {
    key: "programId",

    /**
     * Public key that identifies the BpfLoader
     */
    get: function get() {
      return new PublicKey('0x8000000000000000000000000000000000000000000000000000000000000000');
    }
  }]);

  return BpfLoader;
}();

/**
 * Represents a condition that is met by executing a `applySignature()`
 * transaction
 *
 * @typedef {Object} SignatureCondition
 * @property {string} type Must equal the string 'timestamp'
 * @property {PublicKey} from Public key from which `applySignature()` will be accepted from
 */

/**
 * @private
 */
function serializePayment(payment) {
  var toData = payment.to.toBuffer();
  var data = Buffer.alloc(8 + toData.length);
  data.writeUInt32LE(payment.amount, 0);
  toData.copy(data, 8);
  return data;
}
/**
 * @private
 */


function serializeDate(when) {
  var data = Buffer.alloc(8 + 20);
  data.writeUInt32LE(20, 0); // size of timestamp as u64

  function iso(date) {
    function pad(number) {
      if (number < 10) {
        return '0' + number;
      }

      return number;
    }

    return date.getUTCFullYear() + '-' + pad(date.getUTCMonth() + 1) + '-' + pad(date.getUTCDate()) + 'T' + pad(date.getUTCHours()) + ':' + pad(date.getUTCMinutes()) + ':' + pad(date.getUTCSeconds()) + 'Z';
  }

  data.write(iso(when), 8);
  return data;
}
/**
 * @private
 */


function serializeCondition(condition) {
  switch (condition.type) {
    case 'timestamp':
      {
        var date = serializeDate(condition.when);
        var from = condition.from.toBuffer();
        var data = Buffer.alloc(4 + date.length + from.length);
        data.writeUInt32LE(0, 0); // Condition enum = Timestamp

        date.copy(data, 4);
        from.copy(data, 4 + date.length);
        return data;
      }

    case 'signature':
      {
        var dataLayout = BufferLayout.struct([BufferLayout.u32('condition'), publicKey('from')]);

        var _from = condition.from.toBuffer();

        var _data = Buffer.alloc(4 + _from.length);

        dataLayout.encode({
          instruction: 1,
          // Signature
          from: _from
        }, _data);
        return _data;
      }

    default:
      throw new Error("Unknown condition type: ".concat(condition.type));
  }
}
/**
 * Factory class for transactions to interact with the Budget program
 */


var BudgetProgram =
/*#__PURE__*/
function () {
  function BudgetProgram() {
    _classCallCheck(this, BudgetProgram);
  }

  _createClass(BudgetProgram, null, [{
    key: "timestampCondition",

    /**
     * Creates a timestamp condition
     */
    value: function timestampCondition(from, when) {
      return {
        type: 'timestamp',
        from: from,
        when: when
      };
    }
    /**
     * Creates a signature condition
     */

  }, {
    key: "signatureCondition",
    value: function signatureCondition(from) {
      return {
        type: 'signature',
        from: from
      };
    }
    /**
     * Generates a transaction that transfers lamports once any of the conditions are met
     */

  }, {
    key: "pay",
    value: function pay(from, program, to, amount) {
      var data = Buffer.alloc(1024);
      var pos = 0;
      data.writeUInt32LE(0, pos); // NewBudget instruction

      pos += 4;

      for (var _len = arguments.length, conditions = new Array(_len > 4 ? _len - 4 : 0), _key = 4; _key < _len; _key++) {
        conditions[_key - 4] = arguments[_key];
      }

      switch (conditions.length) {
        case 0:
          data.writeUInt32LE(0, pos); // Budget enum = Pay

          pos += 4;
          {
            var payment = serializePayment({
              amount: amount,
              to: to
            });
            payment.copy(data, pos);
            pos += payment.length;
          }
          return new Transaction().add({
            keys: [{
              pubkey: from,
              isSigner: true
            }, {
              pubkey: to,
              isSigner: false
            }],
            programId: this.programId,
            data: data.slice(0, pos)
          });

        case 1:
          data.writeUInt32LE(1, pos); // Budget enum = After

          pos += 4;
          {
            var condition = conditions[0];
            var conditionData = serializeCondition(condition);
            conditionData.copy(data, pos);
            pos += conditionData.length;
            var paymentData = serializePayment({
              amount: amount,
              to: to
            });
            paymentData.copy(data, pos);
            pos += paymentData.length;
          }
          return new Transaction().add({
            keys: [{
              pubkey: from,
              isSigner: true
            }, {
              pubkey: program,
              isSigner: false
            }, {
              pubkey: to,
              isSigner: false
            }],
            programId: this.programId,
            data: data.slice(0, pos)
          });

        case 2:
          data.writeUInt32LE(2, pos); // Budget enum = Or

          pos += 4;
          var _arr = conditions;

          for (var _i = 0; _i < _arr.length; _i++) {
            var _condition = _arr[_i];

            var _conditionData = serializeCondition(_condition);

            _conditionData.copy(data, pos);

            pos += _conditionData.length;

            var _paymentData = serializePayment({
              amount: amount,
              to: to
            });

            _paymentData.copy(data, pos);

            pos += _paymentData.length;
          }

          return new Transaction().add({
            keys: [{
              pubkey: from,
              isSigner: true
            }, {
              pubkey: program,
              isSigner: false
            }, {
              pubkey: to,
              isSigner: false
            }],
            programId: this.programId,
            data: data.slice(0, pos)
          });

        default:
          throw new Error("A maximum of two conditions are support: ".concat(conditions.length, " provided"));
      }
    }
    /**
     * Generates a transaction that transfers lamports once both conditions are met
     */

  }, {
    key: "payOnBoth",
    value: function payOnBoth(from, program, to, amount, condition1, condition2) {
      var data = Buffer.alloc(1024);
      var pos = 0;
      data.writeUInt32LE(0, pos); // NewBudget instruction

      pos += 4;
      data.writeUInt32LE(3, pos); // Budget enum = And

      pos += 4;
      var _arr2 = [condition1, condition2];

      for (var _i2 = 0; _i2 < _arr2.length; _i2++) {
        var condition = _arr2[_i2];
        var conditionData = serializeCondition(condition);
        conditionData.copy(data, pos);
        pos += conditionData.length;
      }

      var paymentData = serializePayment({
        amount: amount,
        to: to
      });
      paymentData.copy(data, pos);
      pos += paymentData.length;
      return new Transaction().add({
        keys: [{
          pubkey: from,
          isSigner: true
        }, {
          pubkey: program,
          isSigner: false
        }, {
          pubkey: to,
          isSigner: false
        }],
        programId: this.programId,
        data: data.slice(0, pos)
      });
    }
    /**
     * Generates a transaction that applies a timestamp, which could enable a
     * pending payment to proceed.
     */

  }, {
    key: "applyTimestamp",
    value: function applyTimestamp(from, program, to, when) {
      var whenData = serializeDate(when);
      var data = Buffer.alloc(4 + whenData.length);
      data.writeUInt32LE(1, 0); // ApplyTimestamp instruction

      whenData.copy(data, 4);
      return new Transaction().add({
        keys: [{
          pubkey: from,
          isSigner: true
        }, {
          pubkey: program,
          isSigner: false
        }, {
          pubkey: to,
          isSigner: false
        }],
        programId: this.programId,
        data: data
      });
    }
    /**
     * Generates a transaction that applies a signature, which could enable a
     * pending payment to proceed.
     */

  }, {
    key: "applySignature",
    value: function applySignature(from, program, to) {
      var dataLayout = BufferLayout.struct([BufferLayout.u32('instruction')]);
      var data = Buffer.alloc(dataLayout.span);
      dataLayout.encode({
        instruction: 2 // ApplySignature instruction

      }, data);
      return new Transaction().add({
        keys: [{
          pubkey: from,
          isSigner: true
        }, {
          pubkey: program,
          isSigner: false
        }, {
          pubkey: to,
          isSigner: false
        }],
        programId: this.programId,
        data: data
      });
    }
  }, {
    key: "programId",

    /**
     * Public key that identifies the Budget program
     */
    get: function get() {
      return new PublicKey('0x8100000000000000000000000000000000000000000000000000000000000000');
    }
    /**
     * The amount of space this program requires
     */

  }, {
    key: "space",
    get: function get() {
      return 128;
    }
  }]);

  return BudgetProgram;
}();

/**
 * Factory class for transactions to interact with a program loader
 */
var NativeLoader =
/*#__PURE__*/
function () {
  function NativeLoader() {
    _classCallCheck(this, NativeLoader);
  }

  _createClass(NativeLoader, null, [{
    key: "load",

    /**
     * Loads a native program
     *
     * @param connection The connection to use
     * @param owner User account to load the program with
     * @param programName Name of the native program
     */
    value: function () {
      var _load = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee(connection, owner, programName) {
        var bytes, programAccount, transaction, loader;
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                bytes = _toConsumableArray(Buffer.from(programName));
                programAccount = new Account(); // Allocate memory for the program account

                transaction = SystemProgram.createAccount(owner.publicKey, programAccount.publicKey, 1 + 1 + 1, bytes.length + 1, NativeLoader.programId);
                _context.next = 5;
                return sendAndConfirmTransaction(connection, transaction, owner);

              case 5:
                loader = new Loader(connection, NativeLoader.programId);
                _context.next = 8;
                return loader.load(programAccount, bytes);

              case 8:
                _context.next = 10;
                return loader.finalize(programAccount);

              case 10:
                return _context.abrupt("return", programAccount.publicKey);

              case 11:
              case "end":
                return _context.stop();
            }
          }
        }, _callee);
      }));

      function load(_x, _x2, _x3) {
        return _load.apply(this, arguments);
      }

      return load;
    }()
  }, {
    key: "programId",

    /**
     * Public key that identifies the NativeLoader
     */
    get: function get() {
      return new PublicKey('0x100000000000000000000000000000000000000000000000000000000000000');
    }
  }]);

  return NativeLoader;
}();

/**
 * Some amount of tokens
 */
var TokenAmount =
/*#__PURE__*/
function (_BN) {
  _inherits(TokenAmount, _BN);

  function TokenAmount() {
    _classCallCheck(this, TokenAmount);

    return _possibleConstructorReturn(this, _getPrototypeOf(TokenAmount).apply(this, arguments));
  }

  _createClass(TokenAmount, [{
    key: "toBuffer",

    /**
     * Convert to Buffer representation
     */
    value: function toBuffer() {
      var a = _get(_getPrototypeOf(TokenAmount.prototype), "toArray", this).call(this).reverse();

      var b = Buffer.from(a);

      if (b.length === 8) {
        return b;
      }

      assert(b.length < 8, 'TokenAmount too large');
      var zeroPad = Buffer.alloc(8);
      b.copy(zeroPad);
      return zeroPad;
    }
    /**
     * Construct a TokenAmount from Buffer representation
     */

  }], [{
    key: "fromBuffer",
    value: function fromBuffer(buffer) {
      assert(buffer.length === 8, "Invalid buffer length: ".concat(buffer.length));
      return new BN(_toConsumableArray(buffer).reverse().map(function (i) {
        return "00".concat(i.toString(16)).slice(-2);
      }).join(''), 16);
    }
  }]);

  return TokenAmount;
}(BN);
/**
 * Information about a token
 */

/**
 * @private
 */
var TokenInfoLayout = BufferLayout.struct([uint64('supply'), BufferLayout.u8('decimals'), rustString('name'), rustString('symbol')]);
/**
 * Information about a token account
 */

/**
 * @private
 */
var TokenAccountInfoLayout = BufferLayout.struct([publicKey('token'), publicKey('owner'), uint64('amount'), BufferLayout.u8('sourceOption'), publicKey('source'), uint64('originalAmount')]);
// This type exists to workaround an esdoc parse error

/**
 * The built-in token program
 */
var SYSTEM_TOKEN_PROGRAM_ID = new PublicKey('0x8300000000000000000000000000000000000000000000000000000000000000');
/**
 * An ERC20-like Token
 */

var Token =
/*#__PURE__*/
function () {
  /**
   * @private
   */

  /**
   * The public key identifying this token
   */

  /**
   * Program Identifier for the Token program
   */

  /**
   * Create a Token object attached to the specific token
   *
   * @param connection The connection to use
   * @param token Public key of the token
   * @param programId Optional token programId, uses the system programId by default
   */
  function Token(connection, token) {
    var programId = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : SYSTEM_TOKEN_PROGRAM_ID;

    _classCallCheck(this, Token);

    _defineProperty(this, "connection", void 0);

    _defineProperty(this, "token", void 0);

    _defineProperty(this, "programId", void 0);

    Object.assign(this, {
      connection: connection,
      token: token,
      programId: programId
    });
  }
  /**
   * Create a new Token
   *
   * @param connection The connection to use
   * @param owner User account that will own the returned Token Account
   * @param supply Total supply of the new token
   * @param name Descriptive name of this token
   * @param symbol Symbol for this token
   * @param decimals Location of the decimal place
   * @param programId Optional token programId, uses the system programId by default
   * @return Token object for the newly minted token, Public key of the Token Account holding the total supply of new tokens
   */


  _createClass(Token, [{
    key: "newAccount",

    /**
     * Create a new and empty token account.
     *
     * This account may then be used as a `transfer()` or `approve()` destination
     *
     * @param owner User account that will own the new token account
     * @param source If not null, create a delegate account that when authorized
     *               may transfer tokens from this `source` account
     * @return Public key of the new empty token account
     */
    value: function () {
      var _newAccount = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee(owner) {
        var source,
            tokenAccount,
            transaction,
            dataLayout,
            data,
            keys,
            _args = arguments;
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                source = _args.length > 1 && _args[1] !== undefined ? _args[1] : null;
                tokenAccount = new Account();
                dataLayout = BufferLayout.struct([BufferLayout.u32('instruction')]);
                data = Buffer.alloc(dataLayout.span);
                dataLayout.encode({
                  instruction: 1 // NewTokenAccount instruction

                }, data); // Allocate memory for the token

                transaction = SystemProgram.createAccount(owner.publicKey, tokenAccount.publicKey, 1, 1 + TokenAccountInfoLayout.span, this.programId);
                _context.next = 8;
                return sendAndConfirmTransaction(this.connection, transaction, owner);

              case 8:
                // Initialize the token account
                keys = [{
                  pubkey: tokenAccount.publicKey,
                  isSigner: true
                }, {
                  pubkey: owner.publicKey,
                  isSigner: false
                }, {
                  pubkey: this.token,
                  isSigner: false
                }];

                if (source) {
                  keys.push({
                    pubkey: source,
                    isSigner: false
                  });
                }

                transaction = new Transaction().add({
                  keys: keys,
                  programId: this.programId,
                  data: data
                });
                _context.next = 13;
                return sendAndConfirmTransaction(this.connection, transaction, tokenAccount);

              case 13:
                return _context.abrupt("return", tokenAccount.publicKey);

              case 14:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function newAccount(_x) {
        return _newAccount.apply(this, arguments);
      }

      return newAccount;
    }()
    /**
     * Retrieve token information
     */

  }, {
    key: "tokenInfo",
    value: function () {
      var _tokenInfo = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee2() {
        var accountInfo, data, tokenInfo;
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _context2.next = 2;
                return this.connection.getAccountInfo(this.token);

              case 2:
                accountInfo = _context2.sent;

                if (accountInfo.owner.equals(this.programId)) {
                  _context2.next = 5;
                  break;
                }

                throw new Error("Invalid token owner: ".concat(JSON.stringify(accountInfo.owner)));

              case 5:
                data = Buffer.from(accountInfo.data);

                if (!(data.readUInt8(0) !== 1)) {
                  _context2.next = 8;
                  break;
                }

                throw new Error("Invalid token data");

              case 8:
                tokenInfo = TokenInfoLayout.decode(data, 1);
                tokenInfo.supply = TokenAmount.fromBuffer(tokenInfo.supply);
                return _context2.abrupt("return", tokenInfo);

              case 11:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function tokenInfo() {
        return _tokenInfo.apply(this, arguments);
      }

      return tokenInfo;
    }()
    /**
     * Retrieve account information
     *
     * @param account Public key of the token account
     */

  }, {
    key: "accountInfo",
    value: function () {
      var _accountInfo = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee3(account) {
        var accountInfo, data, tokenAccountInfo;
        return _regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                _context3.next = 2;
                return this.connection.getAccountInfo(account);

              case 2:
                accountInfo = _context3.sent;

                if (accountInfo.owner.equals(this.programId)) {
                  _context3.next = 5;
                  break;
                }

                throw new Error("Invalid token account owner");

              case 5:
                data = Buffer.from(accountInfo.data);

                if (!(data.readUInt8(0) !== 2)) {
                  _context3.next = 8;
                  break;
                }

                throw new Error("Invalid token account data");

              case 8:
                tokenAccountInfo = TokenAccountInfoLayout.decode(data, 1);
                tokenAccountInfo.token = new PublicKey(tokenAccountInfo.token);
                tokenAccountInfo.owner = new PublicKey(tokenAccountInfo.owner);
                tokenAccountInfo.amount = TokenAmount.fromBuffer(tokenAccountInfo.amount);

                if (tokenAccountInfo.sourceOption === 0) {
                  tokenAccountInfo.source = null;
                  tokenAccountInfo.originalAmount = new TokenAmount();
                } else {
                  tokenAccountInfo.source = new PublicKey(tokenAccountInfo.source);
                  tokenAccountInfo.originalAmount = TokenAmount.fromBuffer(tokenAccountInfo.originalAmount);
                }

                if (tokenAccountInfo.token.equals(this.token)) {
                  _context3.next = 15;
                  break;
                }

                throw new Error("Invalid token account token: ".concat(JSON.stringify(tokenAccountInfo.token), " !== ").concat(JSON.stringify(this.token)));

              case 15:
                return _context3.abrupt("return", tokenAccountInfo);

              case 16:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function accountInfo(_x2) {
        return _accountInfo.apply(this, arguments);
      }

      return accountInfo;
    }()
    /**
     * Transfer tokens to another account
     *
     * @param owner Owner of the source token account
     * @param source Source token account
     * @param destination Destination token account
     * @param amount Number of tokens to transfer
     */

  }, {
    key: "transfer",
    value: function () {
      var _transfer = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee4(owner, source, destination, amount) {
        return _regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                _context4.t0 = sendAndConfirmTransaction;
                _context4.t1 = this.connection;
                _context4.t2 = new Transaction();
                _context4.next = 5;
                return this.transferInstruction(owner.publicKey, source, destination, amount);

              case 5:
                _context4.t3 = _context4.sent;
                _context4.t4 = _context4.t2.add.call(_context4.t2, _context4.t3);
                _context4.t5 = owner;
                _context4.next = 10;
                return (0, _context4.t0)(_context4.t1, _context4.t4, _context4.t5);

              case 10:
                return _context4.abrupt("return", _context4.sent);

              case 11:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function transfer(_x3, _x4, _x5, _x6) {
        return _transfer.apply(this, arguments);
      }

      return transfer;
    }()
    /**
     * Grant a third-party permission to transfer up the specified number of tokens from an account
     *
     * @param owner Owner of the source token account
     * @param account Public key of the token account
     * @param delegate Token account authorized to perform a transfer tokens from the source account
     * @param amount Maximum number of tokens the delegate may transfer
     */

  }, {
    key: "approve",
    value: function () {
      var _approve = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee5(owner, account, delegate, amount) {
        return _regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                _context5.next = 2;
                return sendAndConfirmTransaction(this.connection, new Transaction().add(this.approveInstruction(owner.publicKey, account, delegate, amount)), owner);

              case 2:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5, this);
      }));

      function approve(_x7, _x8, _x9, _x10) {
        return _approve.apply(this, arguments);
      }

      return approve;
    }()
    /**
     * Remove approval for the transfer of any remaining tokens
     *
     * @param owner Owner of the source token account
     * @param account Public key of the token account
     * @param delegate Token account to revoke authorization from
     */

  }, {
    key: "revoke",
    value: function revoke(owner, account, delegate) {
      return this.approve(owner, account, delegate, 0);
    }
    /**
     * Assign a new owner to the account
     *
     * @param owner Owner of the token account
     * @param account Public key of the token account
     * @param newOwner New owner of the token account
     */

  }, {
    key: "setOwner",
    value: function () {
      var _setOwner = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee6(owner, account, newOwner) {
        return _regeneratorRuntime.wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                _context6.next = 2;
                return sendAndConfirmTransaction(this.connection, new Transaction().add(this.setOwnerInstruction(owner.publicKey, account, newOwner)), owner);

              case 2:
              case "end":
                return _context6.stop();
            }
          }
        }, _callee6, this);
      }));

      function setOwner(_x11, _x12, _x13) {
        return _setOwner.apply(this, arguments);
      }

      return setOwner;
    }()
    /**
     * Construct a Transfer instruction
     *
     * @param owner Owner of the source token account
     * @param source Source token account
     * @param destination Destination token account
     * @param amount Number of tokens to transfer
     */

  }, {
    key: "transferInstruction",
    value: function () {
      var _transferInstruction = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee7(owner, source, destination, amount) {
        var accountInfo, dataLayout, data, keys;
        return _regeneratorRuntime.wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                _context7.next = 2;
                return this.accountInfo(source);

              case 2:
                accountInfo = _context7.sent;

                if (owner.equals(accountInfo.owner)) {
                  _context7.next = 5;
                  break;
                }

                throw new Error('Account owner mismatch');

              case 5:
                dataLayout = BufferLayout.struct([BufferLayout.u32('instruction'), uint64('amount')]);
                data = Buffer.alloc(dataLayout.span);
                dataLayout.encode({
                  instruction: 2,
                  // Transfer instruction
                  amount: new TokenAmount(amount).toBuffer()
                }, data);
                keys = [{
                  pubkey: owner,
                  isSigner: true
                }, {
                  pubkey: source,
                  isSigner: false
                }, {
                  pubkey: destination,
                  isSigner: false
                }];

                if (accountInfo.source) {
                  keys.push({
                    pubkey: accountInfo.source,
                    isSigner: false
                  });
                }

                return _context7.abrupt("return", new TransactionInstruction({
                  keys: keys,
                  programId: this.programId,
                  data: data
                }));

              case 11:
              case "end":
                return _context7.stop();
            }
          }
        }, _callee7, this);
      }));

      function transferInstruction(_x14, _x15, _x16, _x17) {
        return _transferInstruction.apply(this, arguments);
      }

      return transferInstruction;
    }()
    /**
     * Construct an Approve instruction
     *
     * @param owner Owner of the source token account
     * @param account Public key of the token account
     * @param delegate Token account authorized to perform a transfer tokens from the source account
     * @param amount Maximum number of tokens the delegate may transfer
     */

  }, {
    key: "approveInstruction",
    value: function approveInstruction(owner, account, delegate, amount) {
      var dataLayout = BufferLayout.struct([BufferLayout.u32('instruction'), uint64('amount')]);
      var data = Buffer.alloc(dataLayout.span);
      dataLayout.encode({
        instruction: 3,
        // Approve instruction
        amount: new TokenAmount(amount).toBuffer()
      }, data);
      return new TransactionInstruction({
        keys: [{
          pubkey: owner,
          isSigner: true
        }, {
          pubkey: account,
          isSigner: false
        }, {
          pubkey: delegate,
          isSigner: false
        }],
        programId: this.programId,
        data: data
      });
    }
    /**
     * Construct an Revoke instruction
     *
     * @param owner Owner of the source token account
     * @param account Public key of the token account
     * @param delegate Token account authorized to perform a transfer tokens from the source account
     */

  }, {
    key: "revokeInstruction",
    value: function revokeInstruction(owner, account, delegate) {
      return this.approveInstruction(owner, account, delegate, 0);
    }
    /**
     * Construct a SetOwner instruction
     *
     * @param owner Owner of the token account
     * @param account Public key of the token account
     * @param newOwner New owner of the token account
     */

  }, {
    key: "setOwnerInstruction",
    value: function setOwnerInstruction(owner, account, newOwner) {
      var dataLayout = BufferLayout.struct([BufferLayout.u32('instruction')]);
      var data = Buffer.alloc(dataLayout.span);
      dataLayout.encode({
        instruction: 4 // SetOwner instruction

      }, data);
      return new TransactionInstruction({
        keys: [{
          pubkey: owner,
          isSigner: true
        }, {
          pubkey: account,
          isSigner: false
        }, {
          pubkey: newOwner,
          isSigner: false
        }],
        programId: this.programId,
        data: data
      });
    }
  }], [{
    key: "createNewToken",
    value: function () {
      var _createNewToken = _asyncToGenerator(
      /*#__PURE__*/
      _regeneratorRuntime.mark(function _callee8(connection, owner, supply, name, symbol, decimals) {
        var programId,
            tokenAccount,
            token,
            initialAccountPublicKey,
            transaction,
            dataLayout,
            data,
            encodeLength,
            _args8 = arguments;
        return _regeneratorRuntime.wrap(function _callee8$(_context8) {
          while (1) {
            switch (_context8.prev = _context8.next) {
              case 0:
                programId = _args8.length > 6 && _args8[6] !== undefined ? _args8[6] : SYSTEM_TOKEN_PROGRAM_ID;
                tokenAccount = new Account();
                token = new Token(connection, tokenAccount.publicKey, programId);
                _context8.next = 5;
                return token.newAccount(owner, null);

              case 5:
                initialAccountPublicKey = _context8.sent;
                dataLayout = BufferLayout.struct([BufferLayout.u32('instruction'), uint64('supply'), BufferLayout.u8('decimals'), rustString('name'), rustString('symbol')]);
                data = Buffer.alloc(1024);
                encodeLength = dataLayout.encode({
                  instruction: 0,
                  // NewToken instruction
                  supply: supply.toBuffer(),
                  decimals: decimals,
                  name: name,
                  symbol: symbol
                }, data);
                data = data.slice(0, encodeLength);
                // Allocate memory for the tokenAccount account
                transaction = SystemProgram.createAccount(owner.publicKey, tokenAccount.publicKey, 1, 1 + data.length, programId);
                _context8.next = 13;
                return sendAndConfirmTransaction(connection, transaction, owner);

              case 13:
                transaction = new Transaction().add({
                  keys: [{
                    pubkey: tokenAccount.publicKey,
                    isSigner: true
                  }, {
                    pubkey: initialAccountPublicKey,
                    isSigner: false
                  }],
                  programId: programId,
                  data: data
                });
                _context8.next = 16;
                return sendAndConfirmTransaction(connection, transaction, tokenAccount);

              case 16:
                return _context8.abrupt("return", [token, initialAccountPublicKey]);

              case 17:
              case "end":
                return _context8.stop();
            }
          }
        }, _callee8);
      }));

      function createNewToken(_x18, _x19, _x20, _x21, _x22, _x23) {
        return _createNewToken.apply(this, arguments);
      }

      return createNewToken;
    }()
  }]);

  return Token;
}();

/**
 * Sign, send and confirm a raw transaction
 */

function sendAndConfirmRawTransaction(_x, _x2) {
  return _sendAndConfirmRawTransaction.apply(this, arguments);
}

function _sendAndConfirmRawTransaction() {
  _sendAndConfirmRawTransaction = _asyncToGenerator(
  /*#__PURE__*/
  _regeneratorRuntime.mark(function _callee(connection, rawTransaction) {
    var start, signature, status, statusRetries, duration;
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            start = Date.now();
            _context.next = 3;
            return connection.sendRawTransaction(rawTransaction);

          case 3:
            signature = _context.sent;
            // Wait up to a couple slots for a confirmation
            status = null;
            statusRetries = 6;

          case 6:
            _context.next = 8;
            return connection.getSignatureStatus(signature);

          case 8:
            status = _context.sent;

            if (!status) {
              _context.next = 11;
              break;
            }

            return _context.abrupt("break", 18);

          case 11:
            _context.next = 13;
            return sleep(500 * DEFAULT_TICKS_PER_SLOT / NUM_TICKS_PER_SECOND);

          case 13:
            if (!(--statusRetries <= 0)) {
              _context.next = 16;
              break;
            }

            duration = (Date.now() - start) / 1000;
            throw new Error("Raw Transaction '".concat(signature, "' was not confirmed in ").concat(duration.toFixed(2), " seconds (").concat(JSON.stringify(status), ")"));

          case 16:
            _context.next = 6;
            break;

          case 18:
            if (!(status && 'Ok' in status)) {
              _context.next = 20;
              break;
            }

            return _context.abrupt("return", signature);

          case 20:
            throw new Error("Raw transaction ".concat(signature, " failed (").concat(JSON.stringify(status), ")"));

          case 21:
          case "end":
            return _context.stop();
        }
      }
    }, _callee);
  }));
  return _sendAndConfirmRawTransaction.apply(this, arguments);
}

var testnetDefaultChannel = "edge";

/**
 * @private
 */

var endpoint = {
  edge: 'https://api.edge.testnet.solana.com',
  beta: 'https://api.beta.testnet.solana.com',
  stable: 'https://api.testnet.solana.com'
};
/**
 * Retrieves the RPC endpoint URL for the specified testnet release
 * channel
 */

function testnetChannelEndpoint(channel) {
  if (!channel) {
    return endpoint[testnetDefaultChannel];
  }

  if (endpoint[channel]) {
    return endpoint[channel];
  }

  throw new Error("Unknown channel: ".concat(channel));
}

exports.Account = Account;
exports.BpfLoader = BpfLoader;
exports.BudgetProgram = BudgetProgram;
exports.Connection = Connection;
exports.Loader = Loader;
exports.NativeLoader = NativeLoader;
exports.PublicKey = PublicKey;
exports.SystemProgram = SystemProgram;
exports.Token = Token;
exports.TokenAmount = TokenAmount;
exports.Transaction = Transaction;
exports.TransactionInstruction = TransactionInstruction;
exports.sendAndConfirmRawTransaction = sendAndConfirmRawTransaction;
exports.sendAndConfirmTransaction = sendAndConfirmTransaction;
exports.testnetChannelEndpoint = testnetChannelEndpoint;
//# sourceMappingURL=index.cjs.js.map
